@@ -19,12 +19,13 @@
 *)
 
 (**
-  Decode from or encode to the Ogg Vorbis compressed audio format; or get informations about an Ogg Vorbis file.
+  * Decode from or encode to the Ogg Vorbis compressed audio format;
+  * or get informations about an Ogg Vorbis file.
+  *
+  * @author Samuel Mimram, Julien Cristau, David Baelde
+  *)
 
-  @author Samuel Mimram, Julien Cristau, David Baelde
-*)
-
-(* $Id: vorbis.mli,v 1.36 2005/04/15 15:32:07 smimram Exp $ *)
+(* $Id: vorbis.mli,v 1.47 2005/05/31 10:55:40 smimram Exp $ *)
 
 (** {1 Exceptions} *)
 
@@ -46,12 +47,33 @@ exception Invalid_sample_freq
 (** The given file could not be opened. *)
 exception Could_not_open_file
 
-(** The given file is not in ogg/vorbis format. *)
-exception Not_a_vorbis_file
+(** Bitstream is not Vorbis data. *)
+exception Not_vorbis
 
-(** There is a hole in the data. *)
+(** Invalid Vorbis bitstream header. *)
+exception Bad_header
+
+(** A read from media returned an error. *)
+exception Read_error
+
+(** Internal logic fault; indicates a bug or heap/stack corruption. *)
+exception Internal_fault
+
+(** Indicates there was an interruption in the data (one of: garbage between
+  * pages, loss of sync followed by recapture, or a corrupt page). *)
 exception Hole_in_data
 
+(** Indicates that an invalid stream section was supplied,
+  * or the requested link is corrupt. *)
+exception Bad_link
+
+(** Invalid Vorbis bitstream header. *)
+exception Version_mismatch
+
+(** An unknown error happened (in fact it should not have happened, please
+  * report). *)
+exception Unknown_error
+
 (** Error while converting utf8. *)
 exception Utf8_failure of string
 
@@ -93,31 +115,58 @@ val create_encoder_opt :
   string option -> string option -> enc_params -> encoder * string
 
 
-(** Encode a wav buffer into ogg. WARNING: the size of the input buffer should be less than or equal to 1024 bytes. *)
+(** Encode a wav buffer into ogg.
+  * WARNING: the size of the input buffer should be less than or equal to 1024 bytes. *)
 val encode_buffer : encoder -> string -> string
 
+(** Same as [encode_buffer] excepting that [encode_buffer_part enc buf ofs len]
+  * only encodes [len] bytes of buffer [buf] starting at offset [ofs] (in
+  * bytes). *)
+val encode_buffer_part : encoder -> string -> int -> int -> string
+
+
 (** {1 Decoding} *)
 
 (** Output wav format parameters for decoders. *)
 type dec_params =
     {
-      channels : int; (** number of channels (WARNING: it isn't taken in account yet) *)
-      sample_freq : int; (** sampling frequency in Hertz *)
-      sample_size : int; (** sample size in bits *)
-      big_endian : bool; (** store data in big endian ? (default is little endian) *)
-      signed : bool; (** sign for PCM output ? (default is signed) *)
+      sample_size : int; (** sample size in bits (should be 8 or 16) *)
+      big_endian : bool; (** store data in big endian ? (typically [false], little-endian) *)
+      signed : bool; (** sign for PCM output ? (typically [true], signed) *)
     }
 
 (** Ogg/vorbis file opened in input (for decoding). *)
 type dec_file
 
-(** Open an ogg/vorbis file for decoding. Can raise: [Could_not_open_file] or [Not_a_vorbis_file]. *)
+(** Open an ogg/vorbis file for decoding. Can raise: [Could_not_open_file] or
+  * [Not_a_vorbis_file]. Raises: [Read_error], [Not_vorbis],
+  * [Version_mismatch], [Bad_header], [Internal_fault].*)
 val open_dec_file : string -> dec_params -> dec_file
 
 (** Same as [open_dec_file] but uses a file descriptor instead of a file name. The file descriptor is closed by [close_dec_file]. *)
 val open_dec_fd : Unix.file_descr -> dec_params -> dec_file
 
-(** [decode dec_file buf ofs len] decodes [len] octets of wav and put them in [buf] starting at position [pos]. It can raise: [Unix.End_of_file] when the whole file has been decoded, [Hole_in_data], [Invalid_parameters] if all the data cannot fit in the buffer starting at the given position. *)
+(** [open_dec_stream read_func seek_func close_func tell_func params] opens a
+  * stream like [open_dec_file] for decoding but callbacks are used to
+  * manipulate the data. [read_func] should return the requested amount of bytes
+  * (or less if it is the end of file), [seek_funk] should return 0 if the seek
+  * was ok or -1 if the stream is not seekable, [close_func] should close the
+  * stream, and [tell_func] should return the current offset or -1 if there is
+  * no notion of offset in the stream. Raises: [Read_error], [Not_vorbis],
+  * [Version_mismatch], [Bad_header], [Internal_fault]. *)
+val open_dec_stream : (int -> string) -> (unit -> int) -> (unit -> unit) -> (unit -> int) -> dec_params -> dec_file
+
+(** [decode dec_file buf ofs len] decodes [len] octets of wav (should be the
+  * same size as [buf], a typical value
+  * for [len] is 4096) and put them in [buf] starting at position
+  * [pos]. The returned value is the actual number of bytes read. [decode] will
+  * decode at most one vorbis packet per invocation, so the value returned will
+  * generally be less than [len].
+  *
+  * @raise Unix.End_of_file when the whole file has been decoded.
+  * @raise Hole_in_data if there was an interruption of the data.
+  * @raise Invalid_parameters if all the data cannot fit in the buffer starting at the given position.
+  *)
 val decode : dec_file -> string -> int -> int -> int
 
 (** Close an ogg/vorbis file opened for decoding. The file descriptor used for the matching [open_dec_file] is closed (I think). *)
@@ -151,7 +200,21 @@ type info = {
   duration : int; (** duration of the file in seconds *)
 }
 
-(** Get the vorbis comments from a file (see [Vorbis.Info.get_comments]). *)
+(** Get the index of the sequential logical bitstream currently being decoded
+  * (incremented at chaining boundaries even for non-seekable streams). For
+  * seekable streams, it represents the actual chaining index within the
+  * physical bitstream. *)
+val get_dec_file_bitstream : dec_file -> int
+
+(** Get the vorbis comments from a vorbis file (see [Info.get_comments]).
+  * The second argument is the number of the logical bistream (the current
+  * bistream is used if it is set to [None]). *)
+val get_dec_file_comments : dec_file -> int option -> string * ((string * string) array)
+
+(** Get informations about a file being decoded. *)
+val get_dec_file_info : dec_file -> info
+
+(** Get the vorbis comments from a file (see [Info.get_comments]). *)
 val get_comments : string -> string * (string * string) array
 
 (** Get the vorbis information from the file header.*)
@@ -177,21 +240,21 @@ end
 module Info : functor (Io : Iofile) ->
 sig
   (** Get the vorbis comments from a vorbis file.
-    The first string is the vendor string, the elements of the array are
-    of the form (field name, field contents), where 'field name' is
-    uppercase.
-  *)
+    * The first string is the vendor string, the elements of the array are
+    * of the form (field name, field contents), where 'field name' is
+    * uppercase.
+    *)
   val get_comments : string -> string * (string*string) array
 
   (**
-     Get the information in the header of a vorbis file with a safe
-     value of duration.
-     @raise Unix.End_of_file if the file is too small (no complete pages)
-  *)
+    * Get the information in the header of a vorbis file with a safe
+    * value of duration.
+    * @raise Unix.End_of_file if the file is too small (no complete pages).
+    *)
   val get_info : string -> info
 
   (** Same as get_info but uses a heuristic to find the duration of
-      the vorbis file without reading every page header. *)
+    * the vorbis file without reading every page header. *)
   val get_heuristical_info : string -> info
 
   (** Get the size of a given file (in bytes). *)
