@@ -19,7 +19,7 @@
    You should have received a copy of the GNU General Public License
    along with Ocaml-vorbis; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
-*/
+   */
 
 /**
  * Libvorbis bindings for OCaml.
@@ -27,17 +27,18 @@
  * @author Samuel Mimram, Julien Cristau, David Baelde
  */
 
-/* $Id: vorbis_stubs.c,v 1.34 2005/03/29 00:20:33 dbaelde Exp $ */
+/* $Id: vorbis_stubs.c,v 1.50 2005/10/16 20:09:11 smimram Exp $ */
 
 /*
-  TODO:
-  - handle errors
-  - advanced encoder options
+TODO:
+- handle errors
+- advanced encoder options
 */
 
 #define CAML_NAME_SPACE
 #include <caml/alloc.h>
 #include <caml/callback.h>
+#include <caml/custom.h>
 #include <caml/fail.h>
 #include <caml/memory.h>
 #include <caml/misc.h>
@@ -53,7 +54,43 @@
 
 #include "utf8.h"
 
-struct myvorbis__encoder_t
+static void raise_err(int ret)
+{
+  switch(ret)
+  {
+    case OV_HOLE:
+      caml_raise_constant(*caml_named_value("vorbis_exn_hole_in_data"));
+
+    case OV_EREAD:
+      caml_raise_constant(*caml_named_value("vorbis_exn_read_error"));
+
+    case OV_EFAULT:
+      caml_raise_constant(*caml_named_value("vorbis_exn_internal_fault"));
+
+    case OV_ENOTVORBIS:
+      caml_raise_constant(*caml_named_value("vorbis_exn_not_vorbis"));
+
+    case OV_EBADHEADER:
+      caml_raise_constant(*caml_named_value("vorbis_exn_bad_header"));
+
+    case OV_EVERSION:
+      caml_raise_constant(*caml_named_value("vorbis_exn_version_mismatch"));
+
+    case OV_EBADLINK:
+      caml_raise_constant(*caml_named_value("vorbis_exn_bad_link"));
+
+    default:
+      caml_raise_constant(*caml_named_value("vorbis_exn_unknown_error"));
+  }
+}
+
+static void check_err(int ret)
+{
+  if (ret < 0)
+    raise_err(ret);
+}
+
+typedef struct myvorbis__encoder_t
 {
   vorbis_info vi;
   vorbis_dsp_state vd;
@@ -69,27 +106,32 @@ struct myvorbis__encoder_t
   int in_samplesize;
   int in_bigendian;
   int serial;
-};
+} myvorbis_encoder_t;
 
-typedef struct myvorbis__encoder_t myvorbis_encoder_t;
+#define Encoder_val(v) (*((myvorbis_encoder_t**)Data_custom_val(v)))
 
-inline static myvorbis_encoder_t* ve_of_block(value block)
+static void finalize_myvorbis_encoder(value v_e)
 {
-  return (myvorbis_encoder_t*)Field(block, 1);
-}
-
-static void finalize_myvorbis_encoder_t(value v_e)
-{
-  myvorbis_encoder_t *ve = ve_of_block(v_e);
+  myvorbis_encoder_t *ve = Encoder_val(v_e);
 
   ogg_stream_clear(&ve->os);
   vorbis_block_clear(&ve->vb);
   vorbis_dsp_clear(&ve->vd);
-  vorbis_info_clear(&ve->vi);
 
+  vorbis_info_clear(&ve->vi);
   free(ve);
 }
 
+static struct custom_operations encoder_ops =
+{
+  "ocaml_vorbis_encoder",
+  finalize_myvorbis_encoder,
+  custom_compare_default,
+  custom_hash_default,
+  custom_serialize_default,
+  custom_deserialize_default
+};
+
 inline static int int_of_option(value v)
 {
   if(Is_block(v))
@@ -108,9 +150,10 @@ inline static char* string_of_option(value v)
 
 #define copy_buffer(b,len,dst) \
   dst = caml_alloc_string(len); \
-  memmove(String_val(dst), b, len);
+memmove(String_val(dst), b, len);
 
-inline static char* strncat_safe(char* dest, char* src, int pos, int len) {
+inline static char* strncat_safe(char* dest, char* src, int pos, int len)
+{
   dest = (char*)realloc((void*)dest, sizeof(char)*(pos+len));
   memcpy(dest+pos, src, len);
   return dest;
@@ -120,20 +163,20 @@ static void add_tag(vorbis_comment *vc, char *name, char *val)
 {
   char *utf8;
   if(utf8_encode(val, &utf8) >= 0)
-    {
-      if(name == NULL)
-	vorbis_comment_add(vc, utf8);
-      else
-	vorbis_comment_add_tag(vc, name, utf8);
-      free(utf8);
-    }
+  {
+    if(name == NULL)
+      vorbis_comment_add(vc, utf8);
+    else
+      vorbis_comment_add_tag(vc, name, utf8);
+    free(utf8);
+  }
   else
     caml_raise_with_arg(*caml_named_value("vorbis_exn_utf8_failure"), caml_copy_string(val));
 }
 
 static void build_comments(vorbis_comment *vc,
-			   char *artist, char *album, char *title, char *tracknum,
-			   char *date, char *genre)
+    char *artist, char *album, char *title, char *tracknum,
+    char *date, char *genre)
 {
   vorbis_comment_init(vc);
 
@@ -153,18 +196,18 @@ static void build_comments(vorbis_comment *vc,
 }
 
 
-CAMLprim value ocaml_vorbis_get_header_frame(myvorbis_encoder_t * ve, value title, value artist, value genre, value date, value album, value tracknum)
+static value ocaml_vorbis_get_header_frame(value encoder, value title, value artist, value genre, value date, value album, value tracknum)
 {
-  CAMLparam5(title, artist, genre, date, album);
-  CAMLxparam1(tracknum);
+  /* CAMLparam5(encoder, title, artist, genre, date);
+  CAMLxparam2(album, tracknum); */
+  CAMLlocal1(s);
 
   char *ret = NULL;
   int ret_len = 0;
-  CAMLlocal1(s);
+  myvorbis_encoder_t *ve = Encoder_val(encoder);
 
   /* Now, build the three header packets and send through to the stream
-     output stage (but defer actual file output until the main encode loop) */
-
+     output stage (but defer actual file output until the main encode loop). */
   {
     ogg_packet header_main;
     ogg_packet header_comments;
@@ -172,11 +215,12 @@ CAMLprim value ocaml_vorbis_get_header_frame(myvorbis_encoder_t * ve, value titl
     vorbis_comment comments;
 
     build_comments(&comments, string_of_option(artist), string_of_option(album), string_of_option(title), string_of_option(tracknum), string_of_option(date), string_of_option(genre));
-    comments.vendor = "ocaml-vorbis";
+    comments.vendor = malloc(strlen("ocaml-vorbis") + 1);
+    strcpy(comments.vendor, "ocaml-vorbis");
 
     /* Build the packets */
     vorbis_analysis_headerout(&ve->vd, &comments,
-			      &header_main, &header_comments, &header_codebooks);
+        &header_main, &header_comments, &header_codebooks);
 
     /* And stream them out */
     ogg_stream_packetin(&ve->os, &header_main);
@@ -184,24 +228,27 @@ CAMLprim value ocaml_vorbis_get_header_frame(myvorbis_encoder_t * ve, value titl
     ogg_stream_packetin(&ve->os, &header_codebooks);
 
     while(ogg_stream_flush(&ve->os, &ve->og))
-      {
-	ret = (char*)realloc(ret, ret_len + (&ve->og)->header_len + (&ve->og)->body_len);
-	memcpy(ret + ret_len, (&ve->og)->header, (&ve->og)->header_len);
-	memcpy(ret + ret_len + (&ve->og)->header_len, (&ve->og)->body, (&ve->og)->body_len);
-	ret_len += (&ve->og)->header_len + (&ve->og)->body_len;
-      }
+    {
+      ret = (char*)realloc(ret, ret_len + (&ve->og)->header_len + (&ve->og)->body_len);
+      memcpy(ret + ret_len, (&ve->og)->header, (&ve->og)->header_len);
+      memcpy(ret + ret_len + (&ve->og)->header_len, (&ve->og)->body, (&ve->og)->body_len);
+      ret_len += (&ve->og)->header_len + (&ve->og)->body_len;
+    }
+
+    vorbis_comment_clear(&comments);
   }
 
   copy_buffer(ret, ret_len, s);
   free(ret);
-  CAMLreturn(s);
+  /* CAMLreturn(s); */
+  return s;
 }
 
 CAMLprim value ocaml_vorbis_create_encoder(value params, value title, value artist, value genre, value date, value album, value tracknum)
 {
   CAMLparam5(params, title, artist, genre, date);
   CAMLxparam2(album, tracknum);
-  CAMLlocal3(encoder,string,block);
+  CAMLlocal2(encoder, block);
 
   myvorbis_encoder_t *ve;
   int quality_set = 0;
@@ -219,11 +266,12 @@ CAMLprim value ocaml_vorbis_create_encoder(value params, value title, value arti
   int in_samplesize = Int_val(Field(params, 9));
   int in_bigendian = Int_val(Field(params, 10));
 
+  /* TODO: in argument? */
   int serialno;
   srand(time(NULL));
   serialno = rand();
 
-  /* Frequency conversion is performed only for this simple cases */
+  /* Frequency conversion is performed only for these simple cases */
   if(in_samplefreq % sample_freq != 0)
     caml_raise_constant(*caml_named_value("vorbis_exn_invalid_sample_freq"));
   if((channels != 1 && channels != 2) || (in_channels != 1 && in_channels != 2))
@@ -231,7 +279,10 @@ CAMLprim value ocaml_vorbis_create_encoder(value params, value title, value arti
   if(in_samplesize != 8 && in_samplesize != 16)
     caml_raise_constant(*caml_named_value("vorbis_exn_invalid_samplesize"));
 
+  /* We need a pointer to a myvorbis_encoder_t which won't move since vorbis seems to want to update some of the fields */
+  encoder = caml_alloc_custom(&encoder_ops, sizeof(myvorbis_encoder_t*), 0, 1);
   ve = malloc(sizeof(myvorbis_encoder_t));
+  Encoder_val(encoder) = ve;
 
   ve->channels = channels;
   ve->in_channels = in_channels;
@@ -249,46 +300,46 @@ CAMLprim value ocaml_vorbis_create_encoder(value params, value title, value arti
   /* Have vorbisenc choose a mode for us */
   vorbis_info_init(&ve->vi);
   if(quality_set > 0)
+  {
+    if (vorbis_encode_setup_vbr(&ve->vi, channels, sample_freq, quality))
     {
-      if (vorbis_encode_setup_vbr(&ve->vi, channels, sample_freq, quality))
-	{
-	  vorbis_info_clear(&ve->vi);
-	  caml_raise_constant(*caml_named_value("vorbis_exn_invalid_quality"));
-	}
-
-      /* Do we have optional hard quality restrictions? */
-      if (max_bitrate > 0 || min_bitrate > 0)
-	{
-	  struct ovectl_ratemanage_arg ai;
-
-	  vorbis_encode_ctl(&ve->vi, OV_ECTL_RATEMANAGE_GET, &ai);
-	  ai.bitrate_hard_min = min_bitrate * 1000;
-	  ai.bitrate_hard_max = max_bitrate * 1000;
-	  ai.management_active = 1;
-	  vorbis_encode_ctl(&ve->vi, OV_ECTL_RATEMANAGE_SET, &ai);
-	}
+      vorbis_info_clear(&ve->vi);
+      caml_raise_constant(*caml_named_value("vorbis_exn_invalid_quality"));
     }
+
+    /* Do we have optional hard quality restrictions? */
+    if (max_bitrate > 0 || min_bitrate > 0)
+    {
+      struct ovectl_ratemanage_arg ai;
+
+      vorbis_encode_ctl(&ve->vi, OV_ECTL_RATEMANAGE_GET, &ai);
+      ai.bitrate_hard_min = min_bitrate * 1000;
+      ai.bitrate_hard_max = max_bitrate * 1000;
+      ai.management_active = 1;
+      vorbis_encode_ctl(&ve->vi, OV_ECTL_RATEMANAGE_SET, &ai);
+    }
+  }
   else
+  {
+    if (vorbis_encode_setup_managed(&ve->vi, channels, sample_freq,
+          max_bitrate>0?max_bitrate*1000:-1,
+          bitrate*1000,
+          min_bitrate>0?min_bitrate*1000:-1))
     {
-      if (vorbis_encode_setup_managed(&ve->vi, channels, sample_freq,
-				      max_bitrate>0?max_bitrate*1000:-1,
-				      bitrate*1000,
-				      min_bitrate>0?min_bitrate*1000:-1))
-	{
-	  vorbis_info_clear(&ve->vi);
-	  caml_raise_constant(*caml_named_value("vorbis_exn_invalid_parameters"));
-	}
+      vorbis_info_clear(&ve->vi);
+      caml_raise_constant(*caml_named_value("vorbis_exn_invalid_parameters"));
     }
+  }
 
   if (managed && bitrate < 0)
-    {
-      vorbis_encode_ctl(&ve->vi, OV_ECTL_RATEMANAGE_AVG, NULL);
-    }
+  {
+    vorbis_encode_ctl(&ve->vi, OV_ECTL_RATEMANAGE_AVG, NULL);
+  }
   else if(!managed)
-    {
-      /* Turn off management entirely (if it was turned on). */
-      vorbis_encode_ctl(&ve->vi, OV_ECTL_RATEMANAGE_SET, NULL);
-    }
+  {
+    /* Turn off management entirely (if it was turned on). */
+    vorbis_encode_ctl(&ve->vi, OV_ECTL_RATEMANAGE_SET, NULL);
+  }
 
   /* TODO: advanced encoder options */
 
@@ -298,37 +349,36 @@ CAMLprim value ocaml_vorbis_create_encoder(value params, value title, value arti
 
   ogg_stream_init(&ve->os, serialno);
 
-  encoder = caml_alloc_final(2, finalize_myvorbis_encoder_t, sizeof(myvorbis_encoder_t), 1024);
-  Store_field(encoder, 1, (value)ve);
-
   block = caml_alloc_tuple(2);
   Store_field(block, 0, encoder);
   /* TODO : clean this */
-  Store_field(block, 1, ocaml_vorbis_get_header_frame(ve, title, artist, genre, date, album, tracknum));
+  Store_field(block, 1, ocaml_vorbis_get_header_frame(encoder, title, artist, genre, date, album, tracknum));
 
   CAMLreturn(block);
 }
 
-CAMLprim value ocaml_vorbis_create_encoder_byte(value* argv, int argc) {
+CAMLprim value ocaml_vorbis_create_encoder_byte(value* argv, int argc)
+{
   return (ocaml_vorbis_create_encoder(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5], argv[6]));
 }
 
 /* Encode audio from buf, return the available ogg/vorbis data */
-CAMLprim value ocaml_vorbis_encode_buffer(value v_e, value _buf)
+CAMLprim value ocaml_vorbis_encode_buffer(value v_e, value _buf, value _ofs, value _len)
 {
-  CAMLparam2(v_e, _buf);
-  myvorbis_encoder_t *ve = ve_of_block(v_e);
-  int len = caml_string_length(_buf);
+  CAMLparam4(v_e, _buf, _ofs, _len);
+  CAMLlocal1(s);
+  myvorbis_encoder_t *ve = Encoder_val(v_e);
+  int len = Int_val(_len);
   char* in = (char*)malloc(len*sizeof(char));
-  memcpy(in,String_val(_buf),len*sizeof(char));
+  memcpy(in, String_val(_buf)+Int_val(_ofs), len*sizeof(char));
+
+  /* TODO: check that ofs + len <= length(buf) */
 
   /* Ask vorbis the adress where to put the new data.
    * TODO some data can be lost when downsampling: with input at 22050Hz
    * and output at 44100Hz, one sample is lost if the input contains an odd
    * number of samples. It shouldn't hurt too much. */
-  int realsamples =
-    (len * 8 * ve->samplefreq) /
-    (ve->in_samplesize * ve->in_channels * ve->in_samplefreq) ;
+  int realsamples = (len * 8 * ve->samplefreq) / (ve->in_samplesize * ve->in_channels * ve->in_samplefreq);
   float** buffer = vorbis_analysis_buffer(&ve->vd, realsamples);
   int i, j, c;
   int downsample = ve->in_samplefreq / ve->samplefreq ;
@@ -342,9 +392,6 @@ CAMLprim value ocaml_vorbis_encode_buffer(value v_e, value _buf)
   int ret_len = 0;
   int eos = 0;
 
-  /* And finally we'll copy it into a caml value */
-  CAMLlocal1(s);
-
   /* For now, we only work on the C side, and let other Caml threads run */
   caml_enter_blocking_section();
 
@@ -393,31 +440,31 @@ CAMLprim value ocaml_vorbis_encode_buffer(value v_e, value _buf)
   /* While we can get enough data from the library to analyse, one
      block at a time... */
   while(vorbis_analysis_blockout(&ve->vd, &ve->vb) == 1)
+  {
+    /* Do the main analysis, creating a packet */
+    vorbis_analysis(&ve->vb, NULL);
+    vorbis_bitrate_addblock(&ve->vb);
+
+    while(vorbis_bitrate_flushpacket(&ve->vd, &ve->op))
     {
-      /* Do the main analysis, creating a packet */
-      vorbis_analysis(&ve->vb, NULL);
-      vorbis_bitrate_addblock(&ve->vb);
+      /* Add packet to bitstream */
+      ogg_stream_packetin(&ve->os, &ve->op);
 
-      while(vorbis_bitrate_flushpacket(&ve->vd, &ve->op))
-	{
-	  /* Add packet to bitstream */
-	  ogg_stream_packetin(&ve->os, &ve->op);
+      /* If we've gone over a page boundary, we can do actual output,
+         so do so (for however many pages are available) */
+      while(!eos)
+      {
+        if(!ogg_stream_pageout(&ve->os, &ve->og)) break;
 
-	  /* If we've gone over a page boundary, we can do actual output,
-	     so do so (for however many pages are available) */
-	  while(!eos)
-	    {
-	      if(!ogg_stream_pageout(&ve->os, &ve->og)) break;
+        ret = (char*)realloc(ret, ret_len + (&ve->og)->header_len + (&ve->og)->body_len);
+        memcpy(ret + ret_len, (&ve->og)->header, (&ve->og)->header_len);
+        memcpy(ret + ret_len + (&ve->og)->header_len, (&ve->og)->body, (&ve->og)->body_len);
+        ret_len += (&ve->og)->header_len + (&ve->og)->body_len;
 
-	      ret = (char*)realloc(ret, ret_len + (&ve->og)->header_len + (&ve->og)->body_len);
-	      memcpy(ret + ret_len, (&ve->og)->header, (&ve->og)->header_len);
-	      memcpy(ret + ret_len + (&ve->og)->header_len, (&ve->og)->body, (&ve->og)->body_len);
-	      ret_len += (&ve->og)->header_len + (&ve->og)->body_len;
-
-	      if(ogg_page_eos(&ve->og)) eos = 1;
-	    }
-	}
+        if(ogg_page_eos(&ve->og)) eos = 1;
+      }
     }
+  }
 
   /* Back to the Caml world */
   caml_leave_blocking_section();
@@ -430,59 +477,137 @@ CAMLprim value ocaml_vorbis_encode_buffer(value v_e, value _buf)
 }
 
 
-struct myvorbis__dec_file_t
+/* This should be malloced since we might want to register *_func as global root. */
+typedef struct myvorbis__dec_file_t
 {
   OggVorbis_File *ovf;
-  int channels;
-  int sample_freq;
   int sample_size;
   int big_endian;
   int sign;
+  int bitstream;
+  value read_func;
+  value seek_func;
+  value close_func;
+  value tell_func;
+} myvorbis_dec_file_t;
+
+#define Decfile_val(v) (*((myvorbis_dec_file_t**)Data_custom_val(v)))
+
+static void finalize_dec_file(value df)
+{
+  free(Decfile_val(df));
+}
+
+static struct custom_operations decfile_ops =
+{
+  "ocaml_vorbis_decfile",
+  finalize_dec_file,
+  custom_compare_default,
+  custom_hash_default,
+  custom_serialize_default,
+  custom_deserialize_default
 };
 
-
-typedef struct myvorbis__dec_file_t myvorbis_dec_file_t;
-
-
-static myvorbis_dec_file_t* dec_file_of_block(value block)
-{
-  return (myvorbis_dec_file_t*)Field(block, 1);
-}
-
-
-static void finalize_myvorbis_dec_file_t(value d_f)
-{
-  myvorbis_dec_file_t *df = dec_file_of_block(d_f);
-
-  free(df);
-}
-
-
 CAMLprim value ocaml_vorbis_open_dec_file(value fd, value params)
 {
   CAMLparam2(fd, params);
   CAMLlocal1(block);
+  int ret;
 
   FILE *in = NULL;
-  myvorbis_dec_file_t *df = (myvorbis_dec_file_t*)malloc(sizeof(myvorbis_dec_file_t));
+  block = caml_alloc_custom(&decfile_ops, sizeof(myvorbis_dec_file_t*), 0, 1);
+  myvorbis_dec_file_t *df = malloc(sizeof(myvorbis_dec_file_t));
+  Decfile_val(block) = df;
 
   if(!(in = fdopen(Int_val(fd), "rb")))
-      caml_raise_constant(*caml_named_value("vorbis_exn_could_not_open_file"));
+    caml_raise_constant(*caml_named_value("vorbis_exn_could_not_open_file"));
   df->ovf = (OggVorbis_File*)malloc(sizeof(OggVorbis_File));
-  df->channels = Int_val(Field(params, 0));
-  df->sample_freq = Int_val(Field(params, 1));
-  df->sample_size = Int_val(Field(params, 2));
-  df->big_endian = Bool_val(Field(params, 3));
-  df->sign = Bool_val(Field(params, 4));
-  if(ov_open(in, df->ovf, NULL, 0) < 0)
-    {
-      fclose(in);
-      free(df->ovf);
-      caml_raise_constant(*caml_named_value("vorbis_exn_not_a_vorbis_file"));
-    }
+  df->sample_size = Int_val(Field(params, 0));
+  df->big_endian = Bool_val(Field(params, 1));
+  df->sign = Bool_val(Field(params, 2));
+  df->bitstream = 0;
+  df->read_func = (value)NULL;
+  df->seek_func = (value)NULL;
+  df->close_func = (value)NULL;
+  df->tell_func = (value)NULL;
+  ret = ov_open(in, df->ovf, NULL, 0);
+  if (ret < 0)
+  {
+    fclose(in);
+    free(df->ovf);
+    raise_err(ret);
+  }
+
+  CAMLreturn(block);
+}
+
+static size_t read_func_cb(void *ptr, size_t size, size_t nmemb, void *datasource)
+{
+  myvorbis_dec_file_t *df = (myvorbis_dec_file_t*)datasource;
+  value buf = caml_callback(df->read_func, Val_int(size*nmemb));
+  int len = caml_string_length(buf);
+  memcpy(ptr, String_val(buf), len);
+  return len;
+}
+
+static int seek_func_cb(void *datasource, ogg_int64_t offset, int whence)
+{
+  myvorbis_dec_file_t *df = (myvorbis_dec_file_t*)datasource;
+  return Int_val(caml_callback(df->seek_func, Val_int(offset)));
+}
+
+static int close_func_cb(void *datasource)
+{
+  myvorbis_dec_file_t *df = (myvorbis_dec_file_t*)datasource;
+  caml_callback(df->close_func, Val_unit);
+  return 0;
+}
+
+static long tell_func_cb(void *datasource)
+{
+  myvorbis_dec_file_t *df = (myvorbis_dec_file_t*)datasource;
+  return Int_val(caml_callback(df->tell_func, Val_unit));
+}
+
+static ov_callbacks callbacks =
+{
+  .read_func = read_func_cb,
+  .seek_func = seek_func_cb,
+  .close_func = close_func_cb,
+  .tell_func = tell_func_cb
+};
+
+CAMLprim value ocaml_vorbis_open_dec_stream(value read_func, value seek_func, value close_func, value tell_func, value params)
+{
+  CAMLparam5(read_func, seek_func, close_func, tell_func, params);
+  CAMLlocal1(block);
+  int ret = 0;
+  myvorbis_dec_file_t *df;
+
+  block = caml_alloc_custom(&decfile_ops, sizeof(myvorbis_dec_file_t*), 0, 1);
+  df = malloc(sizeof(myvorbis_dec_file_t));
+  Decfile_val(block) = df;
+
+  df->ovf = (OggVorbis_File*)malloc(sizeof(OggVorbis_File));
+  df->sample_size = Int_val(Field(params, 0));
+  df->big_endian = Bool_val(Field(params, 1));
+  df->sign = Bool_val(Field(params, 2));
+  df->bitstream = 0;
+  df->read_func = read_func;
+  caml_register_global_root(&df->read_func);
+  df->seek_func = seek_func;
+  caml_register_global_root(&df->seek_func);
+  df->close_func = close_func;
+  caml_register_global_root(&df->close_func);
+  df->tell_func = tell_func;
+  caml_register_global_root(&df->tell_func);
+  ret = ov_open_callbacks(df, df->ovf, NULL, 0, callbacks);
+  if(ret < 0)
+  {
+    free(df->ovf);
+    raise_err(ret);
+  }
 
-  block = caml_alloc_final(2, finalize_myvorbis_dec_file_t, sizeof(myvorbis_dec_file_t), 1000);
-  Store_field(block, 1, (value)df);
   CAMLreturn(block);
 }
 
@@ -491,20 +616,27 @@ CAMLprim value ocaml_vorbis_decode(value d_f, value buf_, value ofs_, value len_
 {
   CAMLparam4(d_f, buf_, ofs_, len_);
 
-  myvorbis_dec_file_t *df = dec_file_of_block(d_f);
+  myvorbis_dec_file_t *df = Decfile_val(d_f);
   int ret = 0;
-  int bs = 0;
   int ofs = Int_val(ofs_);
   int len = Int_val(len_);
-  char* buf = String_val(buf_) + ofs;
+  char* buf = malloc(len);
 
   if (ofs + len > caml_string_length(buf_))
     caml_raise_constant(*caml_named_value("vorbis_exn_invalid_parameters"));
-  ret = ov_read(df->ovf, buf, len, df->big_endian, df->sample_size / 8, df->sign, &bs);
-  if(ret == 0)
-    caml_raise_end_of_file ();
-  if(ret < 0)
-    caml_raise_constant(*caml_named_value("vorbis_exn_hole_in_data"));
+  ret = ov_read(df->ovf, buf, len, df->big_endian, df->sample_size / 8, df->sign, &df->bitstream);
+  if (ret == 0)
+  {
+    free(buf);
+    caml_raise_end_of_file();
+  }
+  if (ret < 0)
+  {
+    free(buf);
+    raise_err(ret);
+  }
+  memcpy(String_val(buf_) + ofs, buf, len);
+  free(buf);
 
   CAMLreturn(Val_int(ret));
 }
@@ -513,9 +645,17 @@ CAMLprim value ocaml_vorbis_decode(value d_f, value buf_, value ofs_, value len_
 CAMLprim value ocaml_vorbis_close_dec_file(value d_f)
 {
   CAMLparam1(d_f);
-  myvorbis_dec_file_t* df = dec_file_of_block(d_f);
+  myvorbis_dec_file_t* df = Decfile_val(d_f);
 
+  if(df->read_func)
+  {
+    caml_remove_global_root(&df->read_func);
+    caml_remove_global_root(&df->seek_func);
+    caml_remove_global_root(&df->close_func);
+    caml_remove_global_root(&df->tell_func);
+  }
   ov_clear(df->ovf);
+  free(df->ovf);
 
   CAMLreturn(Val_unit);
 }
@@ -523,11 +663,15 @@ CAMLprim value ocaml_vorbis_close_dec_file(value d_f)
 CAMLprim value ocaml_vorbis_utf8_decode(value string)
 {
   CAMLparam1(string);
+  CAMLlocal1(ret);
+
   char* utf8;
-  if(utf8_decode(String_val(string),&utf8) >= 0)
-    {
-      CAMLreturn(caml_copy_string(utf8));
-    }
+  if(utf8_decode(String_val(string), &utf8) >= 0)
+  {
+    ret = caml_copy_string(utf8);
+    free(utf8);
+    CAMLreturn(ret);
+  }
   caml_raise_with_arg(*caml_named_value("vorbis_exn_utf8_failure"), string);
 }
 
@@ -535,30 +679,30 @@ CAMLprim value ocaml_vorbis_encoder_reset(value v_e, value title, value artist,
   CAMLparam5(v_e, title, artist, genre, date);
   CAMLxparam2(album, tracknum);
 
-  myvorbis_encoder_t *s = ve_of_block(v_e);
+  myvorbis_encoder_t *s = Encoder_val(v_e);
 
   ogg_packet op;
   ogg_page og;
   CAMLlocal1(ans);
-  char* ret=NULL;
+  char* ret = NULL;
   int size = 0;
 
   /* tell libvorbis that this is the end of the stream */
   vorbis_analysis_wrote(&s->vd, 0);
 
   while(vorbis_analysis_blockout(&s->vd, &s->vb)==1)
-    {
-      vorbis_analysis(&s->vb, NULL);
-      vorbis_bitrate_addblock(&s->vb);
-      while(vorbis_bitrate_flushpacket(&s->vd, &op))
-	ogg_stream_packetin(&s->os, &op);
-    }
+  {
+    vorbis_analysis(&s->vb, NULL);
+    vorbis_bitrate_addblock(&s->vb);
+    while(vorbis_bitrate_flushpacket(&s->vd, &op))
+      ogg_stream_packetin(&s->os, &op);
+  }
 
   /* flush the last pages in a buffer */
   while(ogg_stream_pageout(&s->os, &og)<=0?0:1) {
-    ret = strncat_safe(ret, og.header, size, og.header_len);
+    ret = strncat_safe(ret, (char*)og.header, size, og.header_len);
     size += og.header_len;
-    ret = strncat_safe(ret, og.body, size, og.body_len);
+    ret = strncat_safe(ret, (char*)og.body, size, og.body_len);
     size += og.body_len;
   }
 
@@ -589,12 +733,12 @@ CAMLprim value ocaml_vorbis_encoder_reset(value v_e, value title, value artist,
 
   /* add the headers to the returned buffer */
   while(ogg_stream_flush(&s->os, &s->og))
-    {
-      ret = (char*)realloc(ret, size + (&s->og)->header_len + (&s->og)->body_len);
-      memcpy(ret + size, (&s->og)->header, (&s->og)->header_len);
-      memcpy(ret + size + (&s->og)->header_len, (&s->og)->body, (&s->og)->body_len);
-      size += (&s->og)->header_len + (&s->og)->body_len;
-    }
+  {
+    ret = (char*)realloc(ret, size + (&s->og)->header_len + (&s->og)->body_len);
+    memcpy(ret + size, (&s->og)->header, (&s->og)->header_len);
+    memcpy(ret + size + (&s->og)->header_len, (&s->og)->body, (&s->og)->body_len);
+    size += (&s->og)->header_len + (&s->og)->body_len;
+  }
 
   copy_buffer(ret, size, ans);
   free(ret);
@@ -605,3 +749,66 @@ value ocaml_vorbis_encoder_reset_byte(value *argv, int argc)
 {
   return (ocaml_vorbis_encoder_reset(argv[0], argv[1], argv[2], argv[3], argv[4], argv[5], argv[6]));
 }
+
+CAMLprim ocaml_vorbis_get_dec_file_bitstream(value d_f)
+{
+  CAMLparam1(d_f);
+  myvorbis_dec_file_t *df = Decfile_val(d_f);
+  CAMLreturn(Val_int(df->bitstream));
+}
+
+CAMLprim ocaml_vorbis_get_dec_file_info(value d_f)
+{
+  CAMLparam1(d_f);
+  myvorbis_dec_file_t *df = Decfile_val(d_f);
+  CAMLlocal1(ans);
+  vorbis_info *vi = df->ovf->vi;
+
+  ans = caml_alloc_tuple(9);
+  Store_field(ans, 0, Val_int(vi->version));
+  Store_field(ans, 1, Val_int(vi->channels));
+  Store_field(ans, 2, Val_int(vi->rate));
+  /* TODO */
+  Store_field(ans, 3, Val_int(0));
+  Store_field(ans, 4, Val_int(0));
+  Store_field(ans, 5, Val_int(0));
+  Store_field(ans, 6, Val_int(0));
+  Store_field(ans, 7, Val_int(0));
+  Store_field(ans, 8, Val_int(0));
+
+  CAMLreturn(ans);
+}
+
+CAMLprim ocaml_vorbis_get_dec_file_comments(value d_f, value link_)
+{
+  CAMLparam2(d_f, link_);
+  myvorbis_dec_file_t *df;
+  int link;
+  int i;
+  vorbis_comment *vc;
+  CAMLlocal2(ans, cmts);
+
+  df = Decfile_val(d_f);
+  if (Is_block(link_))
+    link = Int_val(Field(link_, 0));
+  else
+    link = -1;
+  vc = ov_comment(df->ovf, link);
+
+  if (!vc)
+  {
+    /* TODO: better error */
+    caml_raise_constant(*caml_named_value("vorbis_exn_unknown_error"));
+  }
+
+  cmts = caml_alloc_tuple(vc->comments);
+  for (i = 0; i < vc->comments; i++)
+  {
+    Store_field(cmts, i, caml_copy_string(vc->user_comments[i]));
+  }
+  ans = caml_alloc_tuple(2);
+  Store_field(ans, 0, caml_copy_string(vc->vendor));
+  Store_field(ans, 1, cmts);
+
+  CAMLreturn(ans);
+}
